/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2013, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.cpp
    \brief    prediction class
*/

#include <memory.h>
#include "TComPrediction.h"
#if SHVC_COMPLEXITY_ASSESSMENT
namespace {

/*
 * \brief Derive an AU-aligned address of a given address in a rounding-down manner.
 *
 * \param iAddr
 * \param iAccessUnit
 *
 * \return AU-aligned address
 */
__inline Int iAlignmentRoundDown( Int iAddr, Int iAccessUnit )
{
  Int iAlignedAddr = iAccessUnit * (iAddr/iAccessUnit);
  if ( iAddr < 0 )
  {
    iAlignedAddr -= (iAddr%iAccessUnit) ? iAccessUnit : 0;
  }
  return iAlignedAddr;
}

/*
 * \brief Derive an AU-aligned address of a given address in a rounding-off manner.
 *
 * \param iAddr
 * \param iAccessUnit
 *
 * \return AU-aligned address
 */
__inline Int iAlignmentRoundUp( Int iAddr, Int iAccessUnit )
{
  return iAccessUnit * (iAddr/iAccessUnit) + ((iAddr%iAccessUnit) ? iAccessUnit : 0);
}

} // end of nameless namespace
#endif //SHVC_COMPLEXITY_ASSESSMENT
//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Constructor / destructor / initialize
// ====================================================================================================================

TComPrediction::TComPrediction()
: m_pLumaRecBuffer(0)
, m_iLumaRecStride(0)
{
  m_piYuvExt = NULL;
#if REF_EL_ILMC
  m_pILMCPred[0] = NULL;
  m_pILMCPred[1] = NULL;
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    m_apcMemAccessCalc[i] = 0;
    m_apcMemAccessCalcBlock[i] = 0; 
	m_apcMemAccessCalcPic[i] = 0;
#if SLD_COMPLEXITY_ASSESSMENT
	m_apcMemAccessCalcILMCBlock[i] = 0;
#endif
	m_apcMemAccessCalcRec[i] = 0; 
  }
 pic_level_calc_done=0;
#if COUNT_MULTS
 g_d_num_mults=0;g_d_num_mults_block=0; g_d_num_mults_pic =0;
#endif
#endif//SHVC_COMPLEXITY_ASSESSMENT
}

TComPrediction::~TComPrediction()
{
  
  delete[] m_piYuvExt;

  m_acYuvPred[0].destroy();
  m_acYuvPred[1].destroy();

  m_cYuvPredTemp.destroy();
#if SHVC_COMPLEXITY_ASSESSMENT
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if ( m_apcMemAccessCalc[i] )
    {
      if ( m_apcMemAccessCalc[i]->getCache() )
      {
        delete m_apcMemAccessCalc[i]->getCache();
      }
      delete m_apcMemAccessCalc[i];
    }
    if ( m_apcMemAccessCalcPic[i] )
    {
      if ( m_apcMemAccessCalcPic[i]->getCache() )
      {
        delete m_apcMemAccessCalcPic[i]->getCache();
      }
      delete m_apcMemAccessCalcPic[i];
    }
    if ( m_apcMemAccessCalcRec[i] )
    {
      if ( m_apcMemAccessCalcRec[i]->getCache() )
      {
        delete m_apcMemAccessCalcRec[i]->getCache();
      }
      delete m_apcMemAccessCalcRec[i];
    }
    if ( m_apcMemAccessCalcBlock[i] )
    {
      if ( m_apcMemAccessCalcBlock[i]->getCache() )
      {
        delete m_apcMemAccessCalcBlock[i]->getCache();
      }
      delete m_apcMemAccessCalcBlock[i];
    }
#if SLD_COMPLEXITY_ASSESSMENT
	if ( m_apcMemAccessCalcILMCBlock[i] )
	{
		if ( m_apcMemAccessCalcILMCBlock[i]->getCache() )
		{
			delete m_apcMemAccessCalcILMCBlock[i]->getCache();
		}
		delete m_apcMemAccessCalcILMCBlock[i];
	}
#endif
  }
#endif //SHVC_COMPLEXITY_ASSESSMENT
  if( m_pLumaRecBuffer )
  {
    delete [] m_pLumaRecBuffer;
  }
  
  Int i, j;
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      m_filteredBlock[i][j].destroy();
    }
    m_filteredBlockTmp[i].destroy();
  }

#if REF_EL_ILMC
  if(m_pILMCPred[0])
  { 
	  m_pILMCPred[0]->destroy();
	  delete m_pILMCPred[0];
	  m_pILMCPred[0] = NULL;
  }
  if(m_pILMCPred[1])
  {
	  m_pILMCPred[1]->destroy();
	  delete m_pILMCPred[1];
	  m_pILMCPred[1] = NULL;
  }
#endif
}

#if REF_EL_ILMC
Void TComPrediction::initTempBuff(Bool bILMCPred)
#else
Void TComPrediction::initTempBuff()
#endif
{
  if( m_piYuvExt == NULL )
  {
    Int extWidth  = MAX_CU_SIZE + 16; 
    Int extHeight = MAX_CU_SIZE + 1;
    Int i, j;
    for (i = 0; i < 4; i++)
    {
      m_filteredBlockTmp[i].create(extWidth, extHeight + 7);
      for (j = 0; j < 4; j++)
      {
        m_filteredBlock[i][j].create(extWidth, extHeight);
      }
    }
    m_iYuvExtHeight  = ((MAX_CU_SIZE + 2) << 4);
    m_iYuvExtStride = ((MAX_CU_SIZE  + 8) << 4);
    m_piYuvExt = new Int[ m_iYuvExtStride * m_iYuvExtHeight ];

    // new structure
    m_acYuvPred[0] .create( MAX_CU_SIZE, MAX_CU_SIZE );
    m_acYuvPred[1] .create( MAX_CU_SIZE, MAX_CU_SIZE );

    m_cYuvPredTemp.create( MAX_CU_SIZE, MAX_CU_SIZE );
  }

#if REF_EL_ILMC
  if(bILMCPred)
  {
	  m_pILMCPred[0] = new TComYuv;    
	  m_pILMCPred[0]->create(g_uiMaxCUWidth, g_uiMaxCUHeight); 
	  m_pILMCPred[1] = new TComYuv;
	  m_pILMCPred[1]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
  }
#endif

  if (m_iLumaRecStride != (MAX_CU_SIZE>>1) + 1)
  {
    m_iLumaRecStride =  (MAX_CU_SIZE>>1) + 1;
    if (!m_pLumaRecBuffer)
    {
      m_pLumaRecBuffer = new Pel[ m_iLumaRecStride * m_iLumaRecStride ];
    }
  }
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

// Function for calculating DC value of the reference samples used in Intra prediction
Pel TComPrediction::predIntraGetPredValDC( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft )
{
  assert(iWidth > 0 && iHeight > 0);
  Int iInd, iSum = 0;
  Pel pDcVal;

  if (bAbove)
  {
    for (iInd = 0;iInd < iWidth;iInd++)
    {
      iSum += pSrc[iInd-iSrcStride];
    }
  }
  if (bLeft)
  {
    for (iInd = 0;iInd < iHeight;iInd++)
    {
      iSum += pSrc[iInd*iSrcStride-1];
    }
  }

  if (bAbove && bLeft)
  {
    pDcVal = (iSum + iWidth) / (iWidth + iHeight);
  }
  else if (bAbove)
  {
    pDcVal = (iSum + iWidth/2) / iWidth;
  }
  else if (bLeft)
  {
    pDcVal = (iSum + iHeight/2) / iHeight;
  }
  else
  {
    pDcVal = pSrc[-1]; // Default DC value already calculated and placed in the prediction array if no neighbors are available
  }
  
  return pDcVal;
}

// Function for deriving the angular Intra predictions

/** Function for deriving the simplified angular intra predictions.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 * \param dirMode the intra prediction mode index
 * \param blkAboveAvailable boolean indication if the block above is available
 * \param blkLeftAvailable boolean indication if the block to the left is available
 *
 * This function derives the prediction samples for the angular mode based on the prediction direction indicated by
 * the prediction mode index. The prediction direction is given by the displacement of the bottom row of the block and
 * the reference row above the block in the case of vertical prediction or displacement of the rightmost column
 * of the block and reference column left from the block in the case of the horizontal prediction. The displacement
 * is signalled at 1/32 pixel accuracy. When projection of the predicted pixel falls inbetween reference samples,
 * the predicted value for the pixel is linearly interpolated from the reference samples. All reference samples are taken
 * from the extended main reference.
 */
Void TComPrediction::xPredIntraAng(Int bitDepth, Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter )
{
  Int k,l;
  Int blkSize        = width;
  Pel* pDst          = rpDst;

  // Map the mode index to main prediction direction and angle
  assert( dirMode > 0 ); //no planar
  Bool modeDC        = dirMode < 2;
  Bool modeHor       = !modeDC && (dirMode < 18);
  Bool modeVer       = !modeDC && !modeHor;
  Int intraPredAngle = modeVer ? (Int)dirMode - VER_IDX : modeHor ? -((Int)dirMode - HOR_IDX) : 0;
  Int absAng         = abs(intraPredAngle);
  Int signAng        = intraPredAngle < 0 ? -1 : 1;

  // Set bitshifts and scale the angle parameter to block size
  Int angTable[9]    = {0,    2,    5,   9,  13,  17,  21,  26,  32};
  Int invAngTable[9] = {0, 4096, 1638, 910, 630, 482, 390, 315, 256}; // (256 * 32) / Angle
  Int invAngle       = invAngTable[absAng];
  absAng             = angTable[absAng];
  intraPredAngle     = signAng * absAng;

  // Do the DC prediction
  if (modeDC)
  {
    Pel dcval = predIntraGetPredValDC(pSrc, srcStride, width, height, blkAboveAvailable, blkLeftAvailable);

    for (k=0;k<blkSize;k++)
    {
      for (l=0;l<blkSize;l++)
      {
        pDst[k*dstStride+l] = dcval;
      }
    }
  }

  // Do angular predictions
  else
  {
    Pel* refMain;
    Pel* refSide;
    Pel  refAbove[2*MAX_CU_SIZE+1];
    Pel  refLeft[2*MAX_CU_SIZE+1];

    // Initialise the Main and Left reference array.
    if (intraPredAngle < 0)
    {
      for (k=0;k<blkSize+1;k++)
      {
        refAbove[k+blkSize-1] = pSrc[k-srcStride-1];
      }
      for (k=0;k<blkSize+1;k++)
      {
        refLeft[k+blkSize-1] = pSrc[(k-1)*srcStride-1];
      }
      refMain = (modeVer ? refAbove : refLeft) + (blkSize-1);
      refSide = (modeVer ? refLeft : refAbove) + (blkSize-1);

      // Extend the Main reference to the left.
      Int invAngleSum    = 128;       // rounding for (shift by 8)
      for (k=-1; k>blkSize*intraPredAngle>>5; k--)
      {
        invAngleSum += invAngle;
        refMain[k] = refSide[invAngleSum>>8];
      }
    }
    else
    {
      for (k=0;k<2*blkSize+1;k++)
      {
        refAbove[k] = pSrc[k-srcStride-1];
      }
      for (k=0;k<2*blkSize+1;k++)
      {
        refLeft[k] = pSrc[(k-1)*srcStride-1];
      }
      refMain = modeVer ? refAbove : refLeft;
      refSide = modeVer ? refLeft  : refAbove;
    }

    if (intraPredAngle == 0)
    {
      for (k=0;k<blkSize;k++)
      {
        for (l=0;l<blkSize;l++)
        {
          pDst[k*dstStride+l] = refMain[l+1];
        }
      }

      if ( bFilter )
      {
        for (k=0;k<blkSize;k++)
        {
          pDst[k*dstStride] = Clip3(0, (1<<bitDepth)-1, pDst[k*dstStride] + (( refSide[k+1] - refSide[0] ) >> 1) );
        }
      }
    }
    else
    {
      Int deltaPos=0;
      Int deltaInt;
      Int deltaFract;
      Int refMainIndex;

      for (k=0;k<blkSize;k++)
      {
        deltaPos += intraPredAngle;
        deltaInt   = deltaPos >> 5;
        deltaFract = deltaPos & (32 - 1);

        if (deltaFract)
        {
          // Do linear filtering
          for (l=0;l<blkSize;l++)
          {
            refMainIndex        = l+deltaInt+1;
            pDst[k*dstStride+l] = (Pel) ( ((32-deltaFract)*refMain[refMainIndex]+deltaFract*refMain[refMainIndex+1]+16) >> 5 );
          }
        }
        else
        {
          // Just copy the integer samples
          for (l=0;l<blkSize;l++)
          {
            pDst[k*dstStride+l] = refMain[l+deltaInt+1];
          }
        }
      }
    }

    // Flip the block if this is the horizontal mode
    if (modeHor)
    {
      Pel  tmp;
      for (k=0;k<blkSize-1;k++)
      {
        for (l=k+1;l<blkSize;l++)
        {
          tmp                 = pDst[k*dstStride+l];
          pDst[k*dstStride+l] = pDst[l*dstStride+k];
          pDst[l*dstStride+k] = tmp;
        }
      }
    }
  }
}

Void TComPrediction::predIntraLumaAng(TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc;

  assert( g_aucConvertToBit[ iWidth ] >= 0 ); //   4x  4
  assert( g_aucConvertToBit[ iWidth ] <= 5 ); // 128x128
  assert( iWidth == iHeight  );

  ptrSrc = pcTComPattern->getPredictorPtr( uiDirMode, g_aucConvertToBit[ iWidth ] + 2, m_piYuvExt );

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  // Create the prediction
  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    if ( (iWidth > 16) || (iHeight > 16) )
    {
      xPredIntraAng(g_bitDepthY, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, false );
    }
    else
    {
      xPredIntraAng(g_bitDepthY, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, true );

      if( (uiDirMode == DC_IDX ) && bAbove && bLeft )
      {
        xDCPredFiltering( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight);
      }
    }
  }
}

// Angular chroma
Void TComPrediction::predIntraChromaAng( Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc = piSrc;

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    // Create the prediction
    xPredIntraAng(g_bitDepthC, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, false );
  }
}

/** Function for checking identical motion.
 * \param TComDataCU* pcCU
 * \param UInt PartAddr
 */
Bool TComPrediction::xCheckIdenticalMotion ( TComDataCU* pcCU, UInt PartAddr )
{
  if( pcCU->getSlice()->isInterB() && !pcCU->getSlice()->getPPS()->getWPBiPred() )
  {
    if( pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr) >= 0 && pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr) >= 0)
    {
      Int RefPOCL0 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr))->getPOC();
      Int RefPOCL1 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_1, pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr))->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcCU->getCUMvField(REF_PIC_LIST_0)->getMv(PartAddr) == pcCU->getCUMvField(REF_PIC_LIST_1)->getMv(PartAddr))
      {
        return true;
      }
    }
  }
  return false;
}

Void TComPrediction::motionCompensation ( TComDataCU* pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  if ( iPartIdx >= 0 )
  {
#if ILMC_BLK_DEC
    assert(pcCU->getILMCBlkDec() == false);
#endif
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
    return;
  }

  for ( iPartIdx = 0; iPartIdx < pcCU->getNumPartInter(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

    if ( eRefPicList != REF_PIC_LIST_X )
    {
#if ILMC_BLK_DEC
      assert(pcCU->getILMCBlkDec() == false);
#endif
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
#if RWTH_COMPLEXITY_DEC_STAT
#if SLD_PIX_COMPLEXITY_ASSESSMENT
		assert(g_curLayerId > 0);
#endif
        TComMv cMv0 = pcCU->getCUMvField( REF_PIC_LIST_0 )->getMv( uiPartAddr );
        Int iRefIdx = pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr );
        assert (iRefIdx >= 0);
        TComPic* pcRefPic = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, iRefIdx);
        if ( pcRefPic->isILR( pcCU->getLayerId() ) )
        {
#if SLD_PIX_COMPLEXITY_ASSESSMENT
          if( !pcRefPic->getIsILMCPic() )
		  {
#endif
          assert( cMv0.getHor() == 0 && cMv0.getVer() == 0 );
          g_nNrPixelsBLPred += iWidth * iHeight;
		  g_nNrPixelsIntra  += iWidth * iHeight;
#if SLD_PIX_COMPLEXITY_ASSESSMENT
		  }
#endif
        }
        else
        {
          UInt iIdx = 2; // Full/half/quarter (0,1,2) pel accuracy
          if (cMv0.getAbsHor() % 4 == 0 && cMv0.getAbsVer() % 4 == 0 )
          {
            iIdx = 0;
          }
          else if ( cMv0.getAbsHor() % 2 == 0 && cMv0.getAbsVer() % 2 == 0 )
          {
            iIdx = 1;
          }
          g_nNrPixelsMVUni[iIdx] += iWidth * iHeight;
        }
#endif
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
#if RWTH_COMPLEXITY_DEC_STAT
#if SLD_PIX_COMPLEXITY_ASSESSMENT
		assert(g_curLayerId > 0);
#endif
        // Get ref indices
        Int iRefIdx[2] = {-1, -1};
        for ( Int iRefList = 0; iRefList < 2; iRefList++ )
        {
          RefPicList eCurRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
          iRefIdx[iRefList] = pcCU->getCUMvField( eCurRefPicList )->getRefIdx( uiPartAddr );

          if ( iRefIdx[iRefList] < 0 )
          {
            continue;
          }
        }

        if ( iRefIdx[0]==-1 || iRefIdx[1]==-1 )
        {
          // Uni pred
          RefPicList eCurRefPicList = iRefIdx[0]==-1 ? REF_PIC_LIST_1 : REF_PIC_LIST_0;
          TComMv cMv = pcCU->getCUMvField( eCurRefPicList )->getMv( uiPartAddr );
          Int iCurRefIdx = iRefIdx[0]==-1 ? iRefIdx[1] : iRefIdx[0];
          assert (iCurRefIdx >= 0);
          TComPic* pcRefPic = pcCU->getSlice()->getRefPic(eCurRefPicList, iCurRefIdx);
          if ( pcRefPic->isILR( pcCU->getLayerId() ) )
          {
#if SLD_PIX_COMPLEXITY_ASSESSMENT
            if( !pcRefPic->getIsILMCPic() )
			{
#endif
            assert( cMv.getHor() == 0 && cMv.getVer() == 0 );
            g_nNrPixelsBLPred += iWidth * iHeight;
			g_nNrPixelsIntra  += iWidth * iHeight;
#if SLD_PIX_COMPLEXITY_ASSESSMENT
			}
#endif
          }
          else
          {
            UInt iIdx = 2; // Full/half/quarter (0,1,2) pel accuracy
            if (cMv.getAbsHor() % 4 == 0 && cMv.getAbsVer() % 4 == 0 )
            {
              iIdx = 0;
            }
            else if ( cMv.getAbsHor() % 2 == 0 && cMv.getAbsVer() % 2 == 0 )
            {
              iIdx = 1;
            }
            g_nNrPixelsMVUni[iIdx] += iWidth * iHeight;
          }
        }
        else
        {
#if SLD_PIX_COMPLEXITY_ASSESSMENT
          Bool isILRRef = false;
#endif
          for ( Int iRefList = 0; iRefList < 2; iRefList++ )
          {
            RefPicList eCurRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
            TComMv cMv = pcCU->getCUMvField( eCurRefPicList )->getMv( uiPartAddr );
            assert (iRefIdx[iRefList] >= 0);
            TComPic* pcRefPic = pcCU->getSlice()->getRefPic(eCurRefPicList, iRefIdx[iRefList]);
            if ( pcRefPic->isILR( pcCU->getLayerId() ) )
            {
#if SLD_PIX_COMPLEXITY_ASSESSMENT
              if( !pcRefPic->getIsILMCPic())
			  {
#endif
              assert( cMv.getHor() == 0 && cMv.getVer() == 0 );
              g_nNrPixelsBLPred += iWidth * iHeight;			  
#if SLD_PIX_COMPLEXITY_ASSESSMENT
			  isILRRef = true;
			  }
#endif
            }
            else
            {
              UInt iIdx = 2; // Full/half/quarter (0,1,2) pel accuracy
              if (cMv.getAbsHor() % 4 == 0 && cMv.getAbsVer() % 4 == 0 )
              {
                iIdx = 0;
              }
              else if ( cMv.getAbsHor() % 2 == 0 && cMv.getAbsVer() % 2 == 0 )
              {
                iIdx = 1;
              }
              g_nNrPixelsMVBi[iIdx] += iWidth * iHeight;
            }
          }
#if SLD_PIX_COMPLEXITY_ASSESSMENT
		  if(isILRRef)
			  g_nNrPixelsIntra  += iWidth * iHeight;
#endif
        }
#endif
      }
    }
  }
  return;
}

Void TComPrediction::xPredInterUni ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);

#if REF_IDX_ME_ZEROMV
  assert( ( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) && cMv.getHor() == 0 && cMv.getVer() == 0 ) || pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) == false );
#endif

#if ILMC_BLK_DEC
  if( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->getIsILMCPic() && pcCU->getILMCBlkDec() )
  {
	  predILMCUniPU( pcCU, uiPartAddr, iWidth, iHeight, rpcYuvPred, bi );
	  return;
  }
#endif

#if SLD_PIX_COMPLEXITY_ASSESSMENT
  if( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) )
	  xCalcBLTransformPix(pcCU, uiPartAddr, iWidth, iHeight);
#endif

  xPredInterLumaBlk  ( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
  xPredInterChromaBlk( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
#if SHVC_COMPLEXITY_ASSESSMENT
#if PICTURE_BASE_IL_PROCESSING
#define PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE   32 // as suggested by TI Picture based IL processing uses 32x32 blocks
  if(pic_level_calc_done==0 && pcCU->getLayerId())
  {
	  int i,j,k,iWidthCTU,iHeightCTU;
	  int PicWidth=pcCU->getPic()->getPicYuvRec()->getWidth();
	  int PicHeight=pcCU->getPic()->getPicYuvRec()->getHeight();
	  for(j=PicHeight; j>0; j-=PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE)
	  {
		  iHeightCTU=min(PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE,j);
		  for(i=PicWidth; i>0; i-=PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE)
		  {
			  iWidthCTU=min(PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE,i);	
			  Int PosX = PicWidth-i;
			  Int PosY = PicHeight-j;

			  initMCMemoryAccessCalculator( pcCU->getSlice()->getBaseColPic(pcCU->getLayerId() - 1), 2 ); // for picture-based up-sampling
				if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
				{
#if COUNT_MULTS
	  // count the number of operations needed for interpoaltion
	g_d_num_mults_pic+=iWidthCTU* (iHeightCTU+LUMA_USF_TAPS-1)* LUMA_USF_TAPS; 
	g_d_num_mults_pic+=iWidthCTU* iHeightCTU* LUMA_USF_TAPS; 
	g_d_num_mults_pic+=2*iWidthCTU/2* (iHeightCTU/2+CHROMA_USF_TAPS-1)* CHROMA_USF_TAPS; 
	g_d_num_mults_pic+=2* iWidthCTU/2* iHeightCTU/2* CHROMA_USF_TAPS; 
#endif
				for ( k=0; k<NUM_MEMORY_ARCHITECTURES; k++ )
					{
						if (m_apcMemAccessCalcPic[k])
						{
							if ((i==PicWidth) || (i-PICTURE_BASE_IL_PROCESSING_BLOCK_SIZE<0)) // picture border --> need block extension
							{
									m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW), int(iHeightCTU*g_d_ScalingFactorH), int(PosX*g_d_ScalingFactorW), int(PosY*g_d_ScalingFactorH), LUMA_USF_TAPS, LUMA_USF_TAPS, false);
									m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW/2), int(iHeightCTU*g_d_ScalingFactorH/2), int(PosX*g_d_ScalingFactorW/2), int(PosY*g_d_ScalingFactorH/2), CHROMA_USF_TAPS, CHROMA_USF_TAPS, true);					
							}
							else //  no block extension in horizontal direction is needed
							{
									m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW), int(iHeightCTU*g_d_ScalingFactorH), int(PosX*g_d_ScalingFactorW), int(PosY*g_d_ScalingFactorH), 1, LUMA_USF_TAPS, false);
									m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW/2), int(iHeightCTU*g_d_ScalingFactorH/2), int(PosX*g_d_ScalingFactorW/2), int(PosY*g_d_ScalingFactorH/2), 1, CHROMA_USF_TAPS, true);							
							}
						}
					}
				}
				else  // SNR case
				{
				for ( k=0; k<NUM_MEMORY_ARCHITECTURES; k++ )
					{
						if (m_apcMemAccessCalcPic[k])
						{
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW), int(iHeightCTU*g_d_ScalingFactorH), int(PosX*g_d_ScalingFactorW), int(PosY*g_d_ScalingFactorH), 1, 1, false);
							m_apcMemAccessCalcPic[k]->xCalcAccessBytes(int(iWidthCTU*g_d_ScalingFactorW/2), int(iHeightCTU*g_d_ScalingFactorH/2), int(PosX*g_d_ScalingFactorW/2), int(PosY*g_d_ScalingFactorH/2), 1, 1, true);
						}
					}
				}
 // write re-sampled signal to the new ref frame ("Picture based RefIdx" only)
 // for Re-sampled signal writing to "inter-layer" ref frame
  initMCMemoryAccessCalculator( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx), 2 ); // here should be "eRefPicList, iRefId" which correspond "inter-layer" reference
  xCalcRefRrameWritingMemoryAccess( iWidthCTU, iHeightCTU,int(PosX), int(PosY));       // but only size of picture is important... so there is no performance differnce if we use not "accurate" eRefPicList, iRefId here
		  }
	  }
	  pic_level_calc_done=1;
  }
#endif
#if SLD_COMPLEXITY_ASSESSMENT
  if(pcCU->getLayerId() && pcCU->getSlice()->getBaseColPic(0)->getSlice(0)->getSliceType() != I_SLICE) 
	  assert(!pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()));
#endif
  if (pcCU->getLayerId() && pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) /* detect inter-layer prediction PU*/)
  {  
	    initMCMemoryAccessCalculator( pcCU->getSlice()->getBaseColPic(pcCU->getLayerId() - 1), 1 ); // for block-base up-sampling
		if (g_d_ScalingFactorW !=1.0 || g_d_ScalingFactorH!=1.0) // !SNR case
		{
			xCalcUSMemoryAccessLuma(pcCU, uiPartAddr, iWidth, iHeight, LUMA_USF_TAPS, LUMA_USF_TAPS);
			xCalcUSMemoryAccessChroma(pcCU, uiPartAddr, iWidth, iHeight, CHROMA_USF_TAPS, CHROMA_USF_TAPS);
#if COUNT_MULTS
	        // count the number of operations needed for interpoaltion
			g_d_num_mults_block+=iWidth* (iHeight+LUMA_USF_TAPS-1)* LUMA_USF_TAPS;
			g_d_num_mults_block+=iWidth* iHeight* LUMA_USF_TAPS;
			g_d_num_mults_block+=2*iWidth/2* (iHeight/2+CHROMA_USF_TAPS-1)* CHROMA_USF_TAPS;
			g_d_num_mults_block+=2* iWidth/2* iHeight/2* CHROMA_USF_TAPS; 
#endif
		}
		else                                                     // SNR case
		{
			xCalcUSMemoryAccessLuma(pcCU, uiPartAddr, iWidth, iHeight, 1, 1);
			xCalcUSMemoryAccessChroma(pcCU, uiPartAddr, iWidth, iHeight, 1, 1);
		}
  }
  else
  {
	  initMCMemoryAccessCalculator( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx), 0 ); // for MC
	  xCalcMCMemoryAccessLuma  ( pcCU, uiPartAddr, eRefPicList, iRefIdx, &cMv, iWidth, iHeight, NTAPS_LUMA, NTAPS_LUMA);
	  xCalcMCMemoryAccessChroma( pcCU, uiPartAddr, eRefPicList, iRefIdx, &cMv, iWidth, iHeight, NTAPS_CHROMA, NTAPS_CHROMA);
  }
#endif //SHVC_COMPLEXITY_ASSESSMENT
}

Void TComPrediction::xPredInterBi ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[2] = {-1, -1};

  for ( Int iRefList = 0; iRefList < 2; iRefList++ )
  {
    RefPicList eRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[iRefList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[iRefList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[iRefList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[iRefList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) || 
           ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
      }
      else
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }  
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred ); 
  }
  else
  {
    xWeightedAverage( &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}

/**
 * \brief Generate motion-compensated luma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterLumaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( partAddr );
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorLuma( ref, refStride, dst, dstStride, width, height, xFrac,       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerLuma( ref, refStride, dst, dstStride, width, height, yFrac, true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_LUMA;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorLuma(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, xFrac, false     );
    m_if.filterVerLuma(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              yFrac, false, !bi);    
  }
}

/**
 * \brief Generate motion-compensated chroma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  
  Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  Pel*    refCr     = refPic->getCrAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( partAddr );
  Pel* dstCr = dstPic->getCrAddr( partAddr );
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_CHROMA;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorChroma(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorChroma(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerChroma(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerChroma(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorChroma(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorChroma(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}

Void TComPrediction::xWeightedAverage( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst )
{
  if( iRefIdx0 >= 0 && iRefIdx1 >= 0 )
  {
    rpcYuvDst->addAvg( pcYuvSrc0, pcYuvSrc1, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 >= 0 && iRefIdx1 <  0 )
  {
    pcYuvSrc0->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 <  0 && iRefIdx1 >= 0 )
  {
    pcYuvSrc1->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
}

// AMVP
Void TComPrediction::getMvPredAMVP( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred )
{
  AMVPInfo* pcAMVPInfo = pcCU->getCUMvField(eRefPicList)->getAMVPInfo();
  if( pcAMVPInfo->iN <= 1 )
  {
    rcMvPred = pcAMVPInfo->m_acMvCand[0];

    pcCU->setMVPIdxSubParts( 0, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    pcCU->setMVPNumSubParts( pcAMVPInfo->iN, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    return;
  }

  assert(pcCU->getMVPIdx(eRefPicList,uiPartAddr) >= 0);
  rcMvPred = pcAMVPInfo->m_acMvCand[pcCU->getMVPIdx(eRefPicList,uiPartAddr)];
  return;
}

/** Function for deriving planar intra prediction.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 *
 * This function derives the prediction samples for planar mode (intra coding).
 */
Void TComPrediction::xPredIntraPlanar( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height )
{
  assert(width == height);

  Int k, l, bottomLeft, topRight;
  Int horPred;
  Int leftColumn[MAX_CU_SIZE], topRow[MAX_CU_SIZE], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
  UInt blkSize = width;
  UInt offset2D = width;
  UInt shift1D = g_aucConvertToBit[ width ] + 2;
  UInt shift2D = shift1D + 1;

  // Get left and above reference column and row
  for(k=0;k<blkSize+1;k++)
  {
    topRow[k] = pSrc[k-srcStride];
    leftColumn[k] = pSrc[k*srcStride-1];
  }

  // Prepare intermediate variables used in interpolation
  bottomLeft = leftColumn[blkSize];
  topRight   = topRow[blkSize];
  for (k=0;k<blkSize;k++)
  {
    bottomRow[k]   = bottomLeft - topRow[k];
    rightColumn[k] = topRight   - leftColumn[k];
    topRow[k]      <<= shift1D;
    leftColumn[k]  <<= shift1D;
  }

  // Generate prediction signal
  for (k=0;k<blkSize;k++)
  {
    horPred = leftColumn[k] + offset2D;
    for (l=0;l<blkSize;l++)
    {
      horPred += rightColumn[k];
      topRow[l] += bottomRow[l];
      rpDst[k*dstStride+l] = ( (horPred + topRow[l]) >> shift2D );
    }
  }
}

/** Function for filtering intra DC predictor.
 * \param pSrc pointer to reconstructed sample array
 * \param iSrcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param iWidth the width of the block
 * \param iHeight the height of the block
 *
 * This function performs filtering left and top edges of the prediction samples for DC mode (intra coding).
 */
Void TComPrediction::xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight )
{
  Pel* pDst = rpDst;
  Int x, y, iDstStride2, iSrcStride2;

  // boundary pixels processing
  pDst[0] = (Pel)((pSrc[-iSrcStride] + pSrc[-1] + 2 * pDst[0] + 2) >> 2);

  for ( x = 1; x < iWidth; x++ )
  {
    pDst[x] = (Pel)((pSrc[x - iSrcStride] +  3 * pDst[x] + 2) >> 2);
  }

  for ( y = 1, iDstStride2 = iDstStride, iSrcStride2 = iSrcStride-1; y < iHeight; y++, iDstStride2+=iDstStride, iSrcStride2+=iSrcStride )
  {
    pDst[iDstStride2] = (Pel)((pSrc[iSrcStride2] + 3 * pDst[iDstStride2] + 2) >> 2);
  }

  return;
}

#if SVC_UPSAMPLING
#if SCALED_REF_LAYER_OFFSETS
Void TComPrediction::upsampleBasePic( UInt refLayerIdc, TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic, const Window window)
{
  m_cUsf.upsampleBasePic( refLayerIdc, pcUsPic, pcBasePic, pcTempPic, window);
}
#else
Void TComPrediction::upsampleBasePic( UInt refLayerIdc, TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic)
{
  m_cUsf.upsampleBasePic( refLayerIdc, pcUsPic, pcBasePic, pcTempPic);
}
#endif
#endif

#if REF_EL_ILMC
Void TComPrediction::predILMCPic(TComPic *pcELPic, TComPic *pcILRPic, TComPic *pcILMCPic, TComList<TComPic*>* pcListPic)
{
	TComSlice *pcSlice          = pcELPic->getSlice(pcELPic->getCurrSliceIdx());
	UInt      uiPicWidth        = pcSlice->getSPS()->getPicWidthInLumaSamples();
	UInt      uiPicHeight       = pcSlice->getSPS()->getPicHeightInLumaSamples();
	//only consider two layer case, needs to modify for multiple layers
	Window    scaleEL           = pcSlice->getSPS()->getScaledRefLayerWindow(0);
	UInt      uiPicWidthOffset  = uiPicWidth  - scaleEL.getWindowLeftOffset() - scaleEL.getWindowRightOffset();
	UInt      uiPicHeightOffset = uiPicHeight - scaleEL.getWindowTopOffset()  - scaleEL.getWindowBottomOffset();

	//only consider two layer case, needs to modify for multiple layers 
	TComPic   *pcPicBL          = pcSlice->getBaseColPic(0);
	UInt      uiPicWidthBL      = pcPicBL->getPicYuvRec()->getWidth();
	UInt      uiPicHeightBL     = pcPicBL->getPicYuvRec()->getHeight();

	assert( uiPicWidthOffset == uiPicWidthBL   || uiPicWidthOffset  == 2*uiPicWidthBL  || 2*uiPicWidthOffset  == 3*uiPicWidthBL );
	assert( uiPicHeightOffset == uiPicHeightBL || uiPicHeightOffset == 2*uiPicHeightBL || 2*uiPicHeightOffset == 3*uiPicHeightBL );

#if BL_MV_COMP
	Int       logBLPUSize = 4;
#else
	Int       logBLPUSize = 2;
#endif
	
	Int       iPUWidth    = ((uiPicWidthOffset<<1)/uiPicWidthBL)<<(logBLPUSize-1);
	Int       iPUHeight   = ((uiPicHeightOffset<<1)/uiPicHeightBL)<<(logBLPUSize-1);
	TComMv    cBlkPos, cBlkSize;

	//each block of ILMC picture
	for (Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += iPUHeight)
	{
		Int iBlkHeight = min(iPUHeight, (Int)(uiPicHeight - iBlkY));
		for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += iPUWidth)
		{
			Int iBlkWidth = min(iPUWidth, (Int)(uiPicWidth - iBlkX));
			cBlkPos.set(iBlkX, iBlkY);
			cBlkSize.set(iBlkWidth, iBlkHeight);
			
			predILMCBlk(pcELPic, pcILRPic, pcILMCPic, &cBlkPos, &cBlkSize, pcListPic);			
		}
	}

	return;
}

Void TComPrediction::predILMCBlk(TComPic *pcELPic, TComPic *pcILRPic, TComPic *pcILMCPic, TComMv *pcBlkPos, TComMv * pcBlkSize, TComList<TComPic*>* pcListPic)
{
	Int iBlkX      = pcBlkPos->getHor();
	Int iBlkY      = pcBlkPos->getVer();
	Int iBlkWidth  = pcBlkSize->getHor();
	Int iBlkHeight = pcBlkSize->getVer();

	TComSlice *pcSlice     = pcELPic->getSlice(pcELPic->getCurrSliceIdx());
	TComSlice *pcSliceILMC = pcILMCPic->getSlice(pcILMCPic->getCurrSliceIdx());

	UInt uiCUAddr, uiAbsPartIdx;
	Int  iRefIdx0 = -1, iRefIdx1 = -1;

	//only consider two layer case, needs to modify for multiple layers
	UInt     refLayerIdc = 0;
	TComDataCU *pcBaseCU = pcILMCPic->getCU(0)->getBaseColCU(refLayerIdc, iBlkX, iBlkY, uiCUAddr, uiAbsPartIdx, false);
	if(pcBaseCU != NULL)
	{
#if BL_MV_COMP
		iRefIdx0             = pcBaseCU->getCUMvFieldCompressed(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
		iRefIdx1             = pcBaseCU->getCUMvFieldCompressed(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);
#else
		iRefIdx0             = pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
		iRefIdx1             = pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);
#endif
	}	

	if(iRefIdx0 >= 0 || iRefIdx1 >= 0)  //inter
	{
		Bool biPred = iRefIdx0 >= 0 && iRefIdx1 >= 0;
		if(biPred)
		{
			Int RefPOCL0 = pcSliceILMC->getRefPOC(REF_PIC_LIST_0, iRefIdx0);
			Int RefPOCL1 = pcSliceILMC->getRefPOC(REF_PIC_LIST_1, iRefIdx1);
			if(RefPOCL0 == RefPOCL1 && pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiAbsPartIdx) == pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiAbsPartIdx)) 
				biPred = false;
		}

		if(!biPred)  //uni pred
		{
			RefPicList list      = iRefIdx0 >= 0 ? REF_PIC_LIST_0 : REF_PIC_LIST_1;
			Int        iRefIdx   = iRefIdx0 >= 0 ? iRefIdx0 : iRefIdx1;

			TComMv     cMvBase   = pcBaseCU->getCUMvField(list)->getMv(uiAbsPartIdx);
			TComMv     cMv       = cMvBase.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );
			TComPic    *RefPic   = pcSlice->xGetRefPic(*pcListPic, pcSliceILMC->getRefPOC((RefPicList)list, iRefIdx));
			TComPicYuv *pcRefPic = RefPic->getPicYuvRec();
			assert(RefPic->getPOC() == pcSliceILMC->getRefPOC((RefPicList)list, iRefIdx));
			assert(RefPic->getLayerId() > 0);

			pcSlice->clipMv(cMv, iBlkX, iBlkY);
			xPredInterLumaBlkByPos  (pcRefPic, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterChromaBlkByPos(pcRefPic, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
		}
		else  //bi pred
		{
			TComMv     cMvBase0       = pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiAbsPartIdx);
			TComMv     cMvBase1       = pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiAbsPartIdx);
			TComMv     cMv0           = cMvBase0.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );
			TComMv     cMv1           = cMvBase1.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );

			TComPic    *RefPic0   = pcSlice->xGetRefPic(*pcListPic, pcSliceILMC->getRefPOC(REF_PIC_LIST_0, iRefIdx0));
			TComPic    *RefPic1   = pcSlice->xGetRefPic(*pcListPic, pcSliceILMC->getRefPOC(REF_PIC_LIST_1, iRefIdx1));
			TComPicYuv *pcRefPic0 = RefPic0->getPicYuvRec();
			TComPicYuv *pcRefPic1 = RefPic1->getPicYuvRec();
			assert(RefPic0->getPOC() == pcSliceILMC->getRefPOC(REF_PIC_LIST_0, iRefIdx0));
			assert(RefPic0->getLayerId() > 0);
			assert(RefPic1->getPOC() == pcSliceILMC->getRefPOC(REF_PIC_LIST_1, iRefIdx1));
			assert(RefPic1->getLayerId() > 0);

			pcSlice->clipMv(cMv0, iBlkX, iBlkY);
			pcSlice->clipMv(cMv1, iBlkX, iBlkY);
			xPredInterLumaBlkByPos  (pcRefPic0, iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterChromaBlkByPos(pcRefPic0, iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterLumaBlkByPos  (pcRefPic1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, m_pILMCPred[1], biPred);
			xPredInterChromaBlkByPos(pcRefPic1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, m_pILMCPred[1], biPred);
			m_pILMCPred[0]->addAvg(m_pILMCPred[1], 0, iBlkWidth, iBlkHeight);
		}

		
		m_pILMCPred[0]->copyBlk(pcILMCPic->getPicYuvRec(),iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#if REF_BL_RESIDUE
		addBlkResiPic(pcELPic->getPicYuvResi(), pcILMCPic->getPicYuvRec(), iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#endif
	}
#if !REF_BL_INTRA_LOOP_FILTER_COND
	else  //intra
	{
		copyBlkPic(pcILRPic->getPicYuvRec(), pcILMCPic->getPicYuvRec(), iBlkX, iBlkY, iBlkWidth, iBlkHeight);
	}
#endif

	return;
}

Void TComPrediction::xPredInterLumaBlkByPos(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
	Int refStride = refPic->getStride();  
	Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
	Pel *ref      = refPic->getLumaAddr() + iBlkY*refStride + iBlkX + refOffset;

	Int dstStride = dstPic->getStride();
	Pel *dst      = dstPic->getLumaAddr( 0 );

	Int xFrac = mv->getHor() & 0x3;
	Int yFrac = mv->getVer() & 0x3;

	if ( yFrac == 0 )
	{
		m_if.filterHorLuma( ref, refStride, dst, dstStride, width, height, xFrac,       !bi );
	}
	else if ( xFrac == 0 )
	{
		m_if.filterVerLuma( ref, refStride, dst, dstStride, width, height, yFrac, true, !bi );
	}
	else
	{
		Int tmpStride = m_filteredBlockTmp[0].getStride();
		Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

		Int filterSize = NTAPS_LUMA;
		Int halfFilterSize = ( filterSize >> 1 );

		m_if.filterHorLuma(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, xFrac, false     );
		m_if.filterVerLuma(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              yFrac, false, !bi);    
	}
}

Void TComPrediction::xPredInterChromaBlkByPos(TComPicYuv *refPic, Int iBlkX, Int iBlkY, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
	Int     refStride  = refPic->getCStride();
	Int     dstStride  = dstPic->getCStride();

	Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;

	Pel*    refCb     = refPic->getCbAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
	Pel*    refCr     = refPic->getCrAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;

	Pel* dstCb = dstPic->getCbAddr( 0 );
	Pel* dstCr = dstPic->getCrAddr( 0 );

	Int     xFrac  = mv->getHor() & 0x7;
	Int     yFrac  = mv->getVer() & 0x7;
	UInt    cxWidth  = width  >> 1;
	UInt    cxHeight = height >> 1;

	Int     extStride = m_filteredBlockTmp[0].getStride();
	Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();

	Int filterSize = NTAPS_CHROMA;

	Int halfFilterSize = (filterSize>>1);

	if ( yFrac == 0 )
	{
		m_if.filterHorChroma(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
		m_if.filterHorChroma(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
	}
	else if ( xFrac == 0 )
	{
		m_if.filterVerChroma(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
		m_if.filterVerChroma(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
	}
	else
	{
		m_if.filterHorChroma(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
		m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);

		m_if.filterHorChroma(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
		m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
	}
}

Void TComPrediction::copyBlkPic(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
	Int iStrideSrc = pcPicYuvSrc->getStride();
	Int iStrideDst = pcPicYuvDst->getStride();
	Pel *pSrc      = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
	Pel *pDst      = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;

	for(Int j=0; j<iBlkHeight; j++)
		memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));

	iStrideSrc = pcPicYuvSrc->getCStride();
	iStrideDst = pcPicYuvDst->getCStride();
	iBlkX >>= 1;
	iBlkY >>= 1;
	iBlkWidth >>=1;
	iBlkHeight >>=1;

	pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
	pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
	for(Int j=0; j<iBlkHeight; j++)
		memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
	pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
	pDst = pcPicYuvDst->getCrAddr() + iBlkY*iStrideDst + iBlkX;
	for(Int j=0; j<iBlkHeight; j++)
		memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));

	return;
}
#endif

#if REF_BL_RESIDUE
Void TComPrediction::addBlkResiPic(TComPicYuv *pcPicYuvResi, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
	Int iStrideResi = pcPicYuvResi->getStride();
	Int iStrideDst  = pcPicYuvDst->getStride();
	Pel *pResi      = pcPicYuvResi->getLumaAddr() + iBlkY*iStrideResi + iBlkX;
	Pel *pDst       = pcPicYuvDst->getLumaAddr()  + iBlkY*iStrideDst  + iBlkX;

	for(Int j = 0; j < iBlkHeight; j++)
	{
		for(Int i = 0; i < iBlkWidth; i++)
			pDst[i] = ClipY(pDst[i] + pResi[i]);

		pDst  += iStrideDst;
		pResi += iStrideResi;
	}

	iStrideResi = pcPicYuvResi->getCStride();
	iStrideDst  = pcPicYuvDst->getCStride();
	iBlkX >>= 1;
	iBlkY >>= 1;
	iBlkWidth >>=1;
	iBlkHeight >>=1;

	pResi = pcPicYuvResi->getCbAddr() + iBlkY*iStrideResi + iBlkX;
	pDst  = pcPicYuvDst->getCbAddr()  + iBlkY*iStrideDst  + iBlkX;
	for(Int j = 0; j < iBlkHeight; j++)
	{
		for(Int i = 0; i < iBlkWidth; i++)
			pDst[i] = ClipC(pDst[i] + pResi[i]);

		pDst  += iStrideDst;
		pResi += iStrideResi;
	}
	pResi = pcPicYuvResi->getCrAddr() + iBlkY*iStrideResi + iBlkX;
	pDst  = pcPicYuvDst->getCrAddr()  + iBlkY*iStrideDst  + iBlkX;
	for(Int j = 0; j < iBlkHeight; j++)
	{
		for(Int i = 0; i < iBlkWidth; i++)
			pDst[i] = ClipC(pDst[i] + pResi[i]);

		pDst  += iStrideDst;
		pResi += iStrideResi;
	}

	return;
}
#endif

#if ILMC_BLK_DEC
Void TComPrediction::predILMCUniPU( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred, Bool bi )
{
	assert(pcCU->getLayerId() > 0);

	TComSlice *pcSlice          = pcCU->getSlice();
	UInt      uiPicWidth        = pcSlice->getSPS()->getPicWidthInLumaSamples();
	UInt      uiPicHeight       = pcSlice->getSPS()->getPicHeightInLumaSamples();
	//only working for two layer case, needs to modify for multiple layers
	Window    scaleEL           = pcSlice->getSPS()->getScaledRefLayerWindow(0);
	UInt      uiPicWidthOffset  = uiPicWidth  - scaleEL.getWindowLeftOffset() - scaleEL.getWindowRightOffset();
	UInt      uiPicHeightOffset = uiPicHeight - scaleEL.getWindowTopOffset()  - scaleEL.getWindowBottomOffset();

	//only working for two layer case, needs to modify for multiple layers
	TComPic   *pcPicBL          = pcSlice->getBaseColPic(0);
	UInt      uiPicWidthBL      = pcPicBL->getPicYuvRec()->getWidth();
	UInt      uiPicHeightBL     = pcPicBL->getPicYuvRec()->getHeight();

	Int       logBLPUSize       = 2;

	Int       iPUWidth          = ((uiPicWidthOffset<<1)/uiPicWidthBL)<<(logBLPUSize-1);
	Int       iPUHeight         = ((uiPicHeightOffset<<1)/uiPicHeightBL)<<(logBLPUSize-1);

	Int       uiLPelX           = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiPartAddr] ];
	Int       uiTPelY           = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiPartAddr] ];

	Int       iX0               = uiLPelX;
	Int       iX1               = uiLPelX + iWidth;
	Int       iY0               = uiTPelY;
	Int       iY1               = uiTPelY + iHeight;

	Int       iBLPUPelX,  iBLPUPelY;
	Int       iBLPUWidth, iBLPUHeight;

	static Char b_DecBlkFlag[16*16];
	memset( b_DecBlkFlag, 0, sizeof(b_DecBlkFlag) );

	for(Int iY = iY0; iY < iY1; iY += iPUHeight)
		for(Int iX = iX0; iX < iX1; iX += iPUWidth)
		{
			Int iHorIdx  = (iX - iX0) >> logBLPUSize;
			Int iVertIdx = (iY - iY0) >> logBLPUSize;

			if(!b_DecBlkFlag[iVertIdx*16 + iHorIdx])
			{
				pcCU->getBLPUPosAndSize(iX, iY, iBLPUPelX, iBLPUPelY, iBLPUWidth, iBLPUHeight);

				assert(iBLPUPelX + iBLPUWidth  > iX);
				assert(iBLPUPelY + iBLPUHeight > iY);

				Int iBlkWidth  = min(iX1, iBLPUPelX + iBLPUWidth) - iX;
				Int iBlkHeight = min(iY1, iBLPUPelY + iBLPUHeight)- iY;

				predILMCUniBlk(pcCU, uiPartAddr, iBlkWidth, iBlkHeight, iX, iY, rpcYuvPred, bi);

				for(Int j = 0; j < (iBlkHeight >> logBLPUSize); j++)
					memset(b_DecBlkFlag + (iVertIdx + j)*16 + iHorIdx, 1, iBlkWidth >> logBLPUSize);
			}
		}

#if SLD_PIX_COMPLEXITY_ASSESSMENT && REF_BL_RESIDUE
		xCalcBLTransformPix(pcCU, uiPartAddr, iWidth, iHeight);		
#endif
	return;
}

#if SLD_PIX_COMPLEXITY_ASSESSMENT
Void TComPrediction::xCalcBLTransformPix(TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight)
{
	assert(pcCU->getLayerId() > 0);

	Int       uiLPelX           = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiPartAddr] ];
	Int       uiTPelY           = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiPartAddr] ];

	Int       iX0               = uiLPelX;
	Int       iX1               = uiLPelX + iWidth;
	Int       iY0               = uiTPelY;
	Int       iY1               = uiTPelY + iHeight;

	for(Int iY = iY0; iY < iY1; iY++)
		for(Int iX = iX0; iX < iX1; iX++)
		{
			UInt uiBLCUAddr, uiBLAbsPartIdx; 
			UInt refLayerIdc = 0;
			TComDataCU *pcBaseCU  = pcCU->getBaseColCU(refLayerIdc, iX, iY, uiBLCUAddr, uiBLAbsPartIdx, false);
			if(pcBaseCU->getPredictionMode(uiBLAbsPartIdx) == MODE_INTRA)
			{
				if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_LUMA) )
					g_nNrPixelsTransform[0] ++;
				if( iY % 2 == 0 && iX % 2 == 0 )
				{
					if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_CHROMA_U) )
						g_nNrPixelsTransform[1] ++;
					if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_CHROMA_V) )
						g_nNrPixelsTransform[2] ++;
				}				
			}
			else
			{
				assert(pcBaseCU->getPredictionMode(uiBLAbsPartIdx) == MODE_INTER);
				if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_LUMA) && pcBaseCU->getCbf(uiBLAbsPartIdx, TEXT_LUMA, pcBaseCU->getTransformIdx(uiBLAbsPartIdx)) )
					g_nNrPixelsTransform[0] ++;
				if( iY % 2 == 0 && iX % 2 == 0)
				{
					if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_CHROMA_U) && pcBaseCU->getCbf(uiBLAbsPartIdx, TEXT_CHROMA_U, pcBaseCU->getTransformIdx(uiBLAbsPartIdx)) )
						g_nNrPixelsTransform[1] ++;
					if( !pcBaseCU->getTransformSkip(uiBLAbsPartIdx, TEXT_CHROMA_V) && pcBaseCU->getCbf(uiBLAbsPartIdx, TEXT_CHROMA_V, pcBaseCU->getTransformIdx(uiBLAbsPartIdx)) )
						g_nNrPixelsTransform[2] ++;
				}				
			}
			
		}

	return;
}
#endif

Void TComPrediction::predILMCUniBlk(TComDataCU* pcCU, UInt uiPartAddr, Int iBlkWidth, Int iBlkHeight, Int iBlkX,  Int iBlkY, TComYuv*& rpcYuvPred, Bool bi)
{
	TComPic   *pcPic       = pcCU->getPic();
	TComSlice *pcSlice     = pcPic->getSlice(pcPic->getCurrSliceIdx());

	UInt uiCUAddr, uiAbsPartIdx;
	Int  iRefIdx0 = -1, iRefIdx1 = -1;

	//only working for two layer case, needs to modify for multiple layers
	UInt       refLayerIdc = 0;
	TComDataCU *pcBaseCU   = pcCU->getBaseColCU(refLayerIdc, iBlkX, iBlkY, uiCUAddr, uiAbsPartIdx, false);
	if(pcBaseCU != NULL)
	{
#if BL_MV_COMP
		iRefIdx0    = pcBaseCU->getCUMvFieldCompressed(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
		iRefIdx1    = pcBaseCU->getCUMvFieldCompressed(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);
#else
		iRefIdx0    = pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(uiAbsPartIdx);
		iRefIdx1    = pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(uiAbsPartIdx);
#endif
	}

	if(iRefIdx0 >= 0 || iRefIdx1 >= 0)  //inter
	{
		Bool biPred = iRefIdx0 >= 0 && iRefIdx1 >= 0;
		if(biPred)
		{
			Int RefPOCL0 = pcBaseCU->getSlice()->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
			Int RefPOCL1 = pcBaseCU->getSlice()->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
			if(RefPOCL0 == RefPOCL1 && pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiAbsPartIdx) == pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiAbsPartIdx)) 
				biPred = false;
		}

		if(!biPred)  //uni pred
		{
			RefPicList list      = iRefIdx0 >= 0 ? REF_PIC_LIST_0 : REF_PIC_LIST_1;
			Int        iRefIdx   = iRefIdx0 >= 0 ? iRefIdx0 : iRefIdx1;

			TComMv     cMvBase   = pcBaseCU->getCUMvField(list)->getMv(uiAbsPartIdx);
			TComMv     cMv       = cMvBase.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );
			Int        refPOC    = pcBaseCU->getSlice()->getRefPic(list, iRefIdx)->getPOC();
			TComPic    *RefPic   = NULL;

			for(Int refIdx = 0; refIdx < pcSlice->getNumRefIdx(list); refIdx++)
				if(pcSlice->getRefPic(list, refIdx)->getPOC() == refPOC)
				{
					RefPic = pcSlice->getRefPic(list, refIdx);
#if SLD_COMPLEXITY_ASSESSMENT
					iRefIdx = refIdx;
#endif
					break;
				}
            assert(RefPic != NULL);
			assert(RefPic->getLayerId() > 0);

			TComPicYuv *pcRefPic = RefPic->getPicYuvRec();

			pcSlice->clipMv(cMv, iBlkX, iBlkY);
			xPredInterLumaBlkByPos  (pcRefPic, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterChromaBlkByPos(pcRefPic, iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
#if REF_BL_RESIDUE
			assert(!pcPic->isSpatialEnhLayer(refLayerIdc));
			m_pILMCPred[0]->addResi(pcSlice->getBaseColPic(refLayerIdc)->getPicYuvResi(), iBlkX, iBlkY, 0, iBlkWidth, iBlkHeight);
#endif
			if(bi)
				m_pILMCPred[0]->applyShiftOffset(0, iBlkWidth, iBlkHeight);
#if SLD_COMPLEXITY_ASSESSMENT
			initMCMemoryAccessCalculator( pcSlice->getRefPic(list, iRefIdx), 4 );
			xCalcILMCMemoryAccessLuma   ( iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, NTAPS_LUMA, NTAPS_LUMA);
			xCalcILMCMemoryAccessChroma ( iBlkX, iBlkY, &cMv, iBlkWidth, iBlkHeight, NTAPS_CHROMA, NTAPS_CHROMA);
#if REF_BL_RESIDUE
			initMCMemoryAccessCalculator( pcSlice->getBaseColPic(0), 4, true);
			xCalcResiMemoryAccessLuma   ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
			xCalcResiMemoryAccessChroma ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#endif
#endif
#if SLD_PIX_COMPLEXITY_ASSESSMENT
			if(!bi)
			{
				UInt iIdx = 2;
				if (cMv.getAbsHor() % 4 == 0 && cMv.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv.getAbsHor() % 2 == 0 && cMv.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVUni[iIdx] += iBlkWidth * iBlkHeight;
			}
			else
			{
				UInt iIdx = 2;
				if (cMv.getAbsHor() % 4 == 0 && cMv.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv.getAbsHor() % 2 == 0 && cMv.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVBi[iIdx] += iBlkWidth * iBlkHeight;
			}
			g_nNrPixelsInter += iBlkWidth * iBlkHeight;
#endif
		}
		else  //bi pred
		{
			TComMv     cMvBase0       = pcBaseCU->getCUMvField(REF_PIC_LIST_0)->getMv(uiAbsPartIdx);
			TComMv     cMvBase1       = pcBaseCU->getCUMvField(REF_PIC_LIST_1)->getMv(uiAbsPartIdx);
			TComMv     cMv0           = cMvBase0.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );
			TComMv     cMv1           = cMvBase1.scaleMv( g_mvScalingFactor[refLayerIdc][0], g_mvScalingFactor[refLayerIdc][1] );

			Int        refPOC0        = pcBaseCU->getSlice()->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
			Int        refPOC1        = pcBaseCU->getSlice()->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
			TComPic    *RefPic0       = NULL;
			TComPic    *RefPic1       = NULL;
			for(Int refIdx = 0; refIdx < pcSlice->getNumRefIdx(REF_PIC_LIST_0); refIdx++)
				if(pcSlice->getRefPic(REF_PIC_LIST_0, refIdx)->getPOC() == refPOC0)
				{
					RefPic0 = pcSlice->getRefPic(REF_PIC_LIST_0, refIdx);
#if SLD_COMPLEXITY_ASSESSMENT
					iRefIdx0 = refIdx;
#endif
					break;
				}
			for(Int refIdx = 0; refIdx < pcSlice->getNumRefIdx(REF_PIC_LIST_1); refIdx++)
				if(pcSlice->getRefPic(REF_PIC_LIST_1, refIdx)->getPOC() == refPOC1)
                {
					RefPic1 = pcSlice->getRefPic(REF_PIC_LIST_1, refIdx);
#if SLD_COMPLEXITY_ASSESSMENT
					iRefIdx1 = refIdx;
#endif
					break;
                }
            assert(RefPic0 != NULL);
			assert(RefPic1 != NULL);
			assert(RefPic0->getLayerId() > 0);
			assert(RefPic1->getLayerId() > 0);
			
			TComPicYuv *pcRefPic0 = RefPic0->getPicYuvRec();
			TComPicYuv *pcRefPic1 = RefPic1->getPicYuvRec();

			pcSlice->clipMv(cMv0, iBlkX, iBlkY);
			pcSlice->clipMv(cMv1, iBlkX, iBlkY);
			xPredInterLumaBlkByPos  (pcRefPic0, iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterChromaBlkByPos(pcRefPic0, iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, m_pILMCPred[0], biPred);
			xPredInterLumaBlkByPos  (pcRefPic1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, m_pILMCPred[1], biPred);
			xPredInterChromaBlkByPos(pcRefPic1, iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, m_pILMCPred[1], biPred);
			m_pILMCPred[0]->addAvg(m_pILMCPred[1], 0, iBlkWidth, iBlkHeight);
#if REF_BL_RESIDUE
			assert(!pcPic->isSpatialEnhLayer(refLayerIdc));
			m_pILMCPred[0]->addResi(pcSlice->getBaseColPic(refLayerIdc)->getPicYuvResi(), iBlkX, iBlkY, 0, iBlkWidth, iBlkHeight);
#endif
			if(bi)
				m_pILMCPred[0]->applyShiftOffset(0, iBlkWidth, iBlkHeight);
#if SLD_COMPLEXITY_ASSESSMENT
			initMCMemoryAccessCalculator( pcSlice->getRefPic(REF_PIC_LIST_0, iRefIdx0), 4 );
			xCalcILMCMemoryAccessLuma   ( iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, NTAPS_LUMA, NTAPS_LUMA);
			xCalcILMCMemoryAccessChroma ( iBlkX, iBlkY, &cMv0, iBlkWidth, iBlkHeight, NTAPS_CHROMA, NTAPS_CHROMA);
			initMCMemoryAccessCalculator( pcSlice->getRefPic(REF_PIC_LIST_1, iRefIdx1), 4 );
			xCalcILMCMemoryAccessLuma   ( iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, NTAPS_LUMA, NTAPS_LUMA);
			xCalcILMCMemoryAccessChroma ( iBlkX, iBlkY, &cMv1, iBlkWidth, iBlkHeight, NTAPS_CHROMA, NTAPS_CHROMA);
#if REF_BL_RESIDUE
			initMCMemoryAccessCalculator( pcSlice->getBaseColPic(0), 4, true);
			xCalcResiMemoryAccessLuma   ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
			xCalcResiMemoryAccessChroma ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#endif
#endif
#if SLD_PIX_COMPLEXITY_ASSESSMENT
			if(!bi)
			{
				UInt iIdx = 2;
				if (cMv0.getAbsHor() % 4 == 0 && cMv0.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv0.getAbsHor() % 2 == 0 && cMv0.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVUni[iIdx] += iBlkWidth * iBlkHeight;

				iIdx = 2;
				if (cMv1.getAbsHor() % 4 == 0 && cMv1.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv1.getAbsHor() % 2 == 0 && cMv1.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVUni[iIdx] += iBlkWidth * iBlkHeight;
			}
			else
			{
				UInt iIdx = 2;
				if (cMv0.getAbsHor() % 4 == 0 && cMv0.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv0.getAbsHor() % 2 == 0 && cMv0.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVBi[iIdx] += iBlkWidth * iBlkHeight;

				iIdx = 2;
				if (cMv1.getAbsHor() % 4 == 0 && cMv1.getAbsVer() % 4 == 0 )
				{
					iIdx = 0;
				}
				else if ( cMv1.getAbsHor() % 2 == 0 && cMv1.getAbsVer() % 2 == 0 )
				{
					iIdx = 1;
				}
				g_nNrPixelsMVBi[iIdx] += iBlkWidth * iBlkHeight;
			}
			g_nNrPixelsInter += iBlkWidth * iBlkHeight;
#endif
		}		
	}
	else  //intra
	{
#if REF_BL_INTRA_LOOP_FILTER_COND
		m_pILMCPred[0]->copyPic2Blk(pcSlice->getILMC()->getPicYuvRec(), iBlkX, iBlkY, 0, 0, iBlkWidth, iBlkHeight);
#else
		m_pILMCPred[0]->copyPic2Blk(pcSlice->getILR()->getPicYuvRec(), iBlkX, iBlkY, 0, 0, iBlkWidth, iBlkHeight);
#endif
		if(bi)
			m_pILMCPred[0]->applyShiftOffset(0, iBlkWidth, iBlkHeight);
#if SLD_COMPLEXITY_ASSESSMENT && REF_BL_RESIDUE
		//copy BL intra block
		initMCMemoryAccessCalculator( pcSlice->getBaseColPic(0), 4);
		xCalcResiMemoryAccessLuma   ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
		xCalcResiMemoryAccessChroma ( iBlkX, iBlkY, iBlkWidth, iBlkHeight);
#endif
#if SLD_PIX_COMPLEXITY_ASSESSMENT
		//g_nNrPixelsMVUni[0] += iBlkWidth * iBlkHeight;
		g_nNrPixelsBLPred += iBlkWidth * iBlkHeight;
		g_nNrPixelsIntra  += iBlkWidth * iBlkHeight;
#endif
	}

	Int         uiLPelXInCU = iBlkX - pcCU->getCUPelX();
	Int         uiTPelYInCU = iBlkY - pcCU->getCUPelY();

	m_pILMCPred[0]->copyBlk(rpcYuvPred, uiLPelXInCU, uiTPelYInCU, iBlkWidth, iBlkHeight);

	return;
}
#endif

#if SHVC_COMPLEXITY_ASSESSMENT 
/*
 * \brief Initialize parameters of an MC memory access computation module.
 *
 * \param pcRefPic
 *
 * \return void
 */

#if SLD_COMPLEXITY_ASSESSMENT
Void TComPrediction::initMCMemoryAccessCalculator( TComPic* pcRefPic, int design_flag, Bool bExt )
#else
Void TComPrediction::initMCMemoryAccessCalculator( TComPic* pcRefPic, int design_flag )
#endif
{
#if SLD_COMPLEXITY_ASSESSMENT
	if(design_flag == 4) 
	{ 
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ ) 
		{ 
			if (m_apcMemAccessCalcILMCBlock[i]) 
			{
				if(bExt)
				{
					m_apcMemAccessCalcILMCBlock[i]->setLumaBitDepth( 16 ); 
					m_apcMemAccessCalcILMCBlock[i]->setChromaBitDepth( 16 );
				}
				else
				{
					m_apcMemAccessCalcILMCBlock[i]->setLumaBitDepth( g_bitDepthY ); 
					m_apcMemAccessCalcILMCBlock[i]->setChromaBitDepth( g_bitDepthC );
				}				 
				m_apcMemAccessCalcILMCBlock[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() ); 
				m_apcMemAccessCalcILMCBlock[i]->setPictureIndex( pcRefPic->getPOC() ); 
			} 
		} 
	}
	else
#endif
	if(design_flag == 3) // for RecSignal writing
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{	
			if (m_apcMemAccessCalcRec[i])
			{
				m_apcMemAccessCalcRec[i]->setLumaBitDepth( g_bitDepthY );
				m_apcMemAccessCalcRec[i]->setChromaBitDepth( g_bitDepthC );
				m_apcMemAccessCalcRec[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcRec[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
	else
	if(design_flag == 2)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalcPic[i])
			{
				m_apcMemAccessCalcPic[i]->setLumaBitDepth( g_bitDepthY );
				m_apcMemAccessCalcPic[i]->setChromaBitDepth( g_bitDepthC );
				m_apcMemAccessCalcPic[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcPic[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
	else if(design_flag == 1)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalcBlock[i])
			{
				m_apcMemAccessCalcBlock[i]->setLumaBitDepth( g_bitDepthY );
				m_apcMemAccessCalcBlock[i]->setChromaBitDepth( g_bitDepthC );
				m_apcMemAccessCalcBlock[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalcBlock[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
	else
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if (m_apcMemAccessCalc[i])
			{
				m_apcMemAccessCalc[i]->setLumaBitDepth( g_bitDepthY );
				m_apcMemAccessCalc[i]->setChromaBitDepth( g_bitDepthC );
				m_apcMemAccessCalc[i]->setPictureWidth( pcRefPic->getPicYuvRec()->getWidth() );
				m_apcMemAccessCalc[i]->setPictureIndex( pcRefPic->getPOC() );
			}
		}
	}
}
Void TComPrediction::xCalcRefRrameWritingMemoryAccess( Int iWidth, Int iHeight, Int iLPelX, Int iTPelY)
{
  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcPic[i]) // needed for Picture based up-sampling only
    {
      m_apcMemAccessCalcPic[i]->xCalcAccessBytes(iWidth, iHeight, iLPelX, iTPelY, 1, 1, false); // we use "1,1" since no block extensio is needed for writing
    }
  }
  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcPic[i]) // needed for Picture based up-sampling only
    {
      m_apcMemAccessCalcPic[i]->xCalcAccessBytes(iWidth>>1, iHeight>>1, iLPelX, iTPelY, 1, 1, true);// we use "1,1" since no block extensio is needed for writing
    }
  }
}
Void TComPrediction::xCalcRecWritingMemoryAccess( Int iWidth, Int iHeight, Int iLPelX, Int iTPelY)
{ 
  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  { 
    if (m_apcMemAccessCalcRec[i])
    {
     m_apcMemAccessCalcRec[i]->xCalcAccessBytes(iWidth, iHeight, iLPelX, iTPelY, 1, 1, false); // we use "1,1" since no block extensio is needed for writing
    }
  }
  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcRec[i])
    {
      m_apcMemAccessCalcRec[i]->xCalcAccessBytes(iWidth>>1, iHeight>>1, iLPelX, iTPelY, 1, 1, true);// we use "1,1" since no block extensio is needed for writing
    }
  }
}
Void TComPrediction::xCalcUSMemoryAccessLuma( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
	/* Derive reference block position and filter tap length for luma. */
	// no more that 2 layers coding is assumed
  Int iLumaPUPosX = pcCU->getSlice()->getFullPelBaseRec(pcCU->getLayerId()-1)->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr );
  Int iLumaPUPosY = pcCU->getSlice()->getFullPelBaseRec(pcCU->getLayerId()-1)->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr );

  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
		m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW), int(iHeight*g_d_ScalingFactorH), int(iLumaPUPosX*g_d_ScalingFactorW), int(iLumaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, false);
    }
  }
}
Void TComPrediction::xCalcUSMemoryAccessChroma( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for chroma. */
  // no more that 2 layers coding is assumed
  Int iChromaPUPosX = pcCU->getSlice()->getFullPelBaseRec(pcCU->getLayerId()-1)->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPUPosY = pcCU->getSlice()->getFullPelBaseRec(pcCU->getLayerId()-1)->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  
  Int iMCTapH = iFilTapH;
  Int iMCTapV = iFilTapV;

  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalcBlock[i])
    {
	    m_apcMemAccessCalcBlock[i]->xCalcAccessBytes(int(iWidth*g_d_ScalingFactorW/2), int(iHeight*g_d_ScalingFactorH/2), int(iChromaPUPosX*g_d_ScalingFactorW), int(iChromaPUPosY*g_d_ScalingFactorH), iMCTapH, iMCTapV, true);
    }
  }
}
/*
 * \brief Compute the amount of MC memory access for a luma block.
 *
 * \param pcCU
 * \param uiPartAddr
 * \param eRefPicList
 * \param iRefIdx
 * \param pcMv
 * \param iWidth
 * \param iHeight
 * \param iFilTapH
 * \param iFilTapV
 *
 * \return void
 */
Void TComPrediction::xCalcMCMemoryAccessLuma( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for luma. */
  Int iLumaPUPosX = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iLumaPUPosY = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr );
  Int iLumaPosX = pcMv->getHor() >> 2;
  Int iLumaPosY = pcMv->getVer() >> 2;
  int ixFrac = (pcMv->getHor() & 0x3);
  int iyFrac = (pcMv->getVer() & 0x3);
  if (ixFrac  == 1 || ixFrac  == 3) iFilTapH=iFilTapH-1;
  if (iyFrac  == 1 || iyFrac  == 3) iFilTapV=iFilTapV-1;

  Int iMCTapH = ((pcMv->getHor() & 0x3) != 0) ? iFilTapH : 0;
  Int iMCTapV = ((pcMv->getVer() & 0x3) != 0) ? iFilTapV : 0;


  /* Calculate memory access bandwidth for luma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalc[i])
    {
      m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
    }
  }
}


/*
 * \brief Compute the amount of MC memory access for a chroma block.
 *
 * \param pcCU
 * \param uiPartAddr
 * \param eRefPicList
 * \param iRefIdx
 * \param pcMv
 * \param iWidth
 * \param iHeight
 * \param iFilTapH
 * \param iFilTapV
 *
 * \return void
 */
Void TComPrediction::xCalcMCMemoryAccessChroma( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV)
{
  /* Derive reference block position and filter tap length for chroma. */
  Int iChromaPUPosX = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosX ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPUPosY = pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec()->getLumaPosY ( pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiPartAddr ) / 2;
  Int iChromaPosX = pcMv->getHor() >> 3;
  Int iChromaPosY = pcMv->getVer() >> 3;
  Int iMCTapH = ((pcMv->getHor() & 0x7) != 0) ? iFilTapH : 0;
  Int iMCTapV = ((pcMv->getVer() & 0x7) != 0) ? iFilTapV : 0;

  /* Calculate memory access bandwidth for chroma. */
  for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
  {
    if (m_apcMemAccessCalc[i])
    {
      m_apcMemAccessCalc[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
    }
  }
}

/*
 * \brief Initialize parameters of an MC memory access computation module.
 *
 * \param pcRefPic
 *
 * \return void
 */

Void TComPrediction::initMCMemoryAccessCalculator( const MemCmpParam& cLumaParam, const MemCmpParam& cChromaParam, int design_flag)
{
#if SLD_COMPLEXITY_ASSESSMENT
	if(design_flag == 4) 
	{ 
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ ) 
		{ 
			{ 
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator; 
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam ); 
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam ); 
				m_apcMemAccessCalcILMCBlock[i] = pMAC; 
			} 
			m_apcMemAccessCalcILMCBlock[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] ); 
			if ( g_aiMemArchCacheType[i] > 0 ) 
			{ 
				TComVirtualCache* pcCache; 
				if ( g_aiMemArchCacheType[i] == 2 ) 
				{ 
					pcCache = new TComVirtualCacheLRU; 
				} 
				else 
				{ 
					pcCache = new TComVirtualCacheFIFO; 
				} 
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES ); 
				m_apcMemAccessCalcILMCBlock[i]->attachCache( pcCache ); 
			} 
		} 
	}
	else
#endif
	if(design_flag == 3)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcRec[i] = pMAC;
			}
			m_apcMemAccessCalcRec[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcRec[i]->attachCache( pcCache );
			}
		}
	} else
	if(design_flag == 2)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcPic[i] = pMAC;
			}
			m_apcMemAccessCalcPic[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcPic[i]->attachCache( pcCache );
			}
		}
	}
	else if(design_flag == 1)
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalcBlock[i] = pMAC;
			}
			m_apcMemAccessCalcBlock[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalcBlock[i]->attachCache( pcCache );
			}
		}
	}
	else
	{
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			{
				TComCompressedFrameMemoryAccessCalculator* pMAC = new TComCompressedFrameMemoryAccessCalculator;
				pMAC->setLumaMemoryCompressionParameterSet( cLumaParam );
				pMAC->setChromaMemoryCompressionParameterSet( cChromaParam );
				m_apcMemAccessCalc[i] = pMAC;
			}
			m_apcMemAccessCalc[i]->initMemoryArchitecture( g_aiMemArchDDRAlignBits[i], g_aiMemArchDDRBurstBits[i] );
			if ( g_aiMemArchCacheType[i] > 0 )
			{
				TComVirtualCache* pcCache;
				if ( g_aiMemArchCacheType[i] == 2 )
				{
					pcCache = new TComVirtualCacheLRU;
				}
				else
				{
					pcCache = new TComVirtualCacheFIFO;
				}
				pcCache->initCache( TComVirtualCache::FOUR_WAY, TComVirtualCache::DEFAULT_LOG2_CACHE_LINE_SIZE, TComVirtualCache::DEFAULT_LOG2_CACHE_NUM_LINES );
				m_apcMemAccessCalc[i]->attachCache( pcCache );
			}
		}
	}
}

/*
 * \brief Update values of an MC memory access computation module.
 *
 * \return void
 */
Void TComPrediction::updatePic( int design_flag )
{
#if SLD_COMPLEXITY_ASSESSMENT
	if (design_flag == 4) 
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ ) 
		{ 
			if ( m_apcMemAccessCalcILMCBlock[i] ) m_apcMemAccessCalcILMCBlock[i]->update(); 
		}
	else
#endif
	if (design_flag == 3)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcRec[i] ) m_apcMemAccessCalcRec[i]->update();
		}
	else
	if (design_flag == 2)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcPic[i] ) m_apcMemAccessCalcPic[i]->update();
		}
	else if (design_flag == 1)
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalcBlock[i] ) m_apcMemAccessCalcBlock[i]->update();
		}
	else
		for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
		{
			if ( m_apcMemAccessCalc[i] ) m_apcMemAccessCalc[i]->update();
		}
}
/*
 * \brief Get the value of total MC memory access measure for a specified memory architecture in byte.
 *
 * \param iIndex Memory archtecture index
 *
 * \return The total MC memory access measure for the specified memory architecture. 
 */

UInt64 TComPrediction::getTotalMCMemoryAccessBytes( Int iIndex, int design_flag  )
{
  assert ( 0 <= iIndex && iIndex < NUM_MEMORY_ARCHITECTURES );
#if SLD_COMPLEXITY_ASSESSMENT
  if (design_flag == 4) 
	  return (m_apcMemAccessCalcILMCBlock[iIndex]) ? m_apcMemAccessCalcILMCBlock[iIndex]->getTotalAccessBytes() : 0;
  else
#endif
  if (design_flag == 3)
	return (m_apcMemAccessCalcRec[iIndex]) ? m_apcMemAccessCalcRec[iIndex]->getTotalAccessBytes() : 0;
  else
  if (design_flag == 2)
	return (m_apcMemAccessCalcPic[iIndex]) ? m_apcMemAccessCalcPic[iIndex]->getTotalAccessBytes() : 0;
  else if (design_flag == 1)
	return (m_apcMemAccessCalcBlock[iIndex]) ? m_apcMemAccessCalcBlock[iIndex]->getTotalAccessBytes() : 0;
  else
	return (m_apcMemAccessCalc[iIndex]) ? m_apcMemAccessCalc[iIndex]->getTotalAccessBytes() : 0;
}

#if SLD_COMPLEXITY_ASSESSMENT
Void TComPrediction::xCalcILMCMemoryAccessLuma( Int iBlkX, Int iBlkY, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV )
{
	/* Derive reference block position and filter tap length for luma. */
	Int iLumaPUPosX = iBlkX;
	Int iLumaPUPosY = iBlkY;
	Int iLumaPosX = pcMv->getHor() >> 2;
	Int iLumaPosY = pcMv->getVer() >> 2;
	int ixFrac = (pcMv->getHor() & 0x3);
	int iyFrac = (pcMv->getVer() & 0x3);
	if (ixFrac  == 1 || ixFrac  == 3) iFilTapH=iFilTapH-1;
	if (iyFrac  == 1 || iyFrac  == 3) iFilTapV=iFilTapV-1;

	Int iMCTapH = ((pcMv->getHor() & 0x3) != 0) ? iFilTapH : 0;
	Int iMCTapV = ((pcMv->getVer() & 0x3) != 0) ? iFilTapV : 0;

	/* Calculate memory access bandwidth for luma. */
	for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
	{
		if (m_apcMemAccessCalcILMCBlock[i])
		{
			m_apcMemAccessCalcILMCBlock[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX+iLumaPosX, iLumaPUPosY+iLumaPosY, iMCTapH, iMCTapV, false);
		}
	}

	return;
}

Void TComPrediction::xCalcILMCMemoryAccessChroma( Int iBlkX, Int iBlkY, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV )
{
	/* Derive reference block position and filter tap length for chroma. */
	Int iChromaPUPosX = iBlkX / 2;
	Int iChromaPUPosY = iBlkY / 2;
	Int iChromaPosX = pcMv->getHor() >> 3;
	Int iChromaPosY = pcMv->getVer() >> 3;
	Int iMCTapH = ((pcMv->getHor() & 0x7) != 0) ? iFilTapH : 0;
	Int iMCTapV = ((pcMv->getVer() & 0x7) != 0) ? iFilTapV : 0;

	/* Calculate memory access bandwidth for chroma. */
	for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
	{
		if (m_apcMemAccessCalcILMCBlock[i])
		{
			m_apcMemAccessCalcILMCBlock[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX+iChromaPosX, iChromaPUPosY+iChromaPosY, iMCTapH, iMCTapV, true);
		}
	}

	return;
}

#if REF_BL_RESIDUE
Void TComPrediction::xCalcResiMemoryAccessLuma( Int iBlkX, Int iBlkY, Int iWidth, Int iHeight )
{
	// Derive reference block position and filter tap length for luma.
	Int iLumaPUPosX = iBlkX;
	Int iLumaPUPosY = iBlkY;

	// Calculate memory access bandwidth for luma.
	for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
	{
		if (m_apcMemAccessCalcILMCBlock[i])
		{
			m_apcMemAccessCalcILMCBlock[i]->xCalcAccessBytes(iWidth, iHeight, iLumaPUPosX, iLumaPUPosY, 1, 1, false);
		}
	}
	return;
}

Void TComPrediction::xCalcResiMemoryAccessChroma( Int iBlkX, Int iBlkY, Int iWidth, Int iHeight )
{
	// Derive reference block position and filter tap length for chroma.
	Int iChromaPUPosX = iBlkX / 2;
	Int iChromaPUPosY = iBlkY / 2;

	// Calculate memory access bandwidth for chroma.
	for ( Int i=0; i<NUM_MEMORY_ARCHITECTURES; i++ )
	{
		if (m_apcMemAccessCalcILMCBlock[i])
		{
			m_apcMemAccessCalcILMCBlock[i]->xCalcAccessBytes(iWidth/2, iHeight/2, iChromaPUPosX, iChromaPUPosY, 1, 1, true);
		}
	}
	return;
}
#endif
#endif
#endif //SHVC_COMPLEXITY_ASSESSMENT
//! \}
